/*
 * Arguments for a Csound task
 *
 * Copyright (c) 2003 by John D. Ramsdell. All rights reserved.
 *
 * This software is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <wx/wxprec.h>

#ifndef WX_PRECOMP
#include <wx/string.h>
#include <wx/event.h>
#include <wx/cmdline.h>
#endif

#include <wx/filename.h>
#include <ctype.h>
#include "args.hpp"

// Some of the arguments handled by this class are not displayed by
// the GUI.  These arguments are passed through unchanged.  They can
// be set by directly editing a Csound Configuration file.

// Args is a model for Settings, which implements the control and the
// view.  Changes to the model are sent to the view via this update
// method.

void Args::update() const
{
  if (m_handler) {
    wxUpdateUIEvent ev(m_id);
    m_handler->AddPendingEvent(ev);
  }
}

Args::Args()
  : m_handler(0), m_id(0)
{
  clear();
}

void Args::clear()
{
  m_orchestra.Clear();
  m_score.Clear();
  m_output.Clear();
  m_log.Clear();
  m_format = NO_FORMAT;
  m_graphics = FULL_GRAPHICS;
  m_output_size = NO_SIZE;
  m_no_peak_chunks = FALSE;
  m_play_after_rendering = FALSE;
  m_verbose = FALSE;
  m_cscore = FALSE;
  m_init_time_only = FALSE;
  m_no_sound = FALSE;
  m_notify = FALSE;
  m_list_opcodes = FALSE;
  m_rewrite_header = FALSE;
  m_message_level.Clear();
  update();
}

void Args::set_handler(wxEvtHandler *handler)
{
  m_handler = handler;
}

void Args::set_id(wxWindowID id)
{
  m_id = id;
}

// Generate a file name relative to a base file name.
inline wxFileName mkrel(const wxFileName& base, const wxFileName& file_name)
{
  if (base.IsOk()) {
    wxFileName copy = file_name;
    copy.MakeRelativeTo(base.GetPath(wxPATH_GET_SEPARATOR));
    return copy;
  }
  else
    return file_name;
}

// Accessors and Modifiers

wxFileName Args::get_orchestra() const
{
  return m_orchestra;
}

wxFileName Args::get_orchestra(const wxFileName& base) const
{
  return mkrel(base, m_orchestra);
}

void Args::set_orchestra(const wxFileName& orchestra)
{
  if (m_orchestra.SameAs(orchestra))
    return;
  m_orchestra = orchestra;
  update();
}

wxFileName Args::get_score() const
{
  return m_score;
}

wxFileName Args::get_score(const wxFileName& base) const
{
  return mkrel(base, m_score);
}

void Args::set_score(const wxFileName& score)
{
  if (m_score.SameAs(score))
    return;
  m_score = score;
  update();
}

wxFileName Args::get_output() const
{
  return m_output;
}

wxFileName Args::get_output(const wxFileName& base) const
{
  return mkrel(base, m_output);
}

void Args::set_output(const wxFileName& output)
{
  if (m_output.SameAs(output))
    return;
  m_output = output;
  update();
}

wxFileName Args::get_log() const
{
  return m_log;
}

wxFileName Args::get_log(const wxFileName& base) const
{
  return mkrel(base, m_log);
}

void Args::set_log(const wxFileName& log)
{
  if (m_log.SameAs(log))
    return;
  m_log = log;
  update();
}

Format Args::get_format() const
{
  return m_format;
}

void Args::set_format(Format format)
{
  if (m_format == format)
    return;
  m_format = format;
  update();
}

Graphics_option Args::get_graphics_option() const
{
  return m_graphics;
}

void Args::set_graphics_option(Graphics_option graphics)
{
  if (m_graphics == graphics)
    return;
  m_graphics = graphics;
  update();
}

Output_size Args::get_output_size() const
{
  return m_output_size;
}

void Args::set_output_size(Output_size output_size)
{
  if (m_output_size == output_size)
    return;
  m_output_size = output_size;
  update();
}

bool Args::get_no_peak_chunks() const
{
  return m_no_peak_chunks;
}

void Args::set_no_peak_chunks(bool no_peak_chunks)
{
  if (m_no_peak_chunks == no_peak_chunks)
    return;
  m_no_peak_chunks = no_peak_chunks;
  update();
}

bool Args::get_play_after_rendering() const
{
  return m_play_after_rendering;
}

void Args::set_play_after_rendering(bool play_after_rendering)
{
  if (m_play_after_rendering == play_after_rendering)
    return;
  m_play_after_rendering = play_after_rendering;
  update();
}

bool Args::get_verbose() const
{
  return m_verbose;
}

void Args::set_verbose(bool verbose)
{
  if (m_verbose == verbose)
    return;
  m_verbose = verbose;
  update();
}

bool Args::get_cscore() const
{
  return m_cscore;
}

void Args::set_cscore(bool cscore)
{
  if (m_cscore == cscore)
    return;
  m_cscore = cscore;
  update();
}

bool Args::get_init_time_only() const
{
  return m_init_time_only;
}

void Args::set_init_time_only(bool init_time_only)
{
  if (m_init_time_only == init_time_only)
    return;
  m_init_time_only = init_time_only;
  update();
}

bool Args::get_no_sound() const
{
  return m_no_sound;
}

void Args::set_no_sound(bool no_sound)
{
  if (m_no_sound == no_sound)
    return;
  m_no_sound = no_sound;
  update();
}

bool Args::get_notify() const
{
  return m_notify;
}

void Args::set_notify(bool notify)
{
  if (m_notify == notify)
    return;
  m_notify = notify;
  update();
}

bool Args::get_list_opcodes() const
{
  return m_list_opcodes;
}

void Args::set_list_opcodes(bool list_opcodes)
{
  if (m_list_opcodes == list_opcodes)
    return;
  m_list_opcodes = list_opcodes;
  update();
}

bool Args::get_rewrite_header() const
{
  return m_rewrite_header;
}

void Args::set_rewrite_header(bool rewrite_header)
{
  if (m_rewrite_header == rewrite_header)
    return;
  m_rewrite_header = rewrite_header;
  update();
}

wxString Args::get_message_level() const
{
  return m_message_level;
}

void Args::set_message_level(const wxString& message_level)
{
  if (m_message_level.IsSameAs(message_level))
    return;
  m_message_level = message_level;
  update();
}

// Generate a command line from the data within.
// File names are relative to a given base file name.
wxArrayString Args::get_command_line(const wxFileName& base)
{
  wxArrayString args;
  args.Add(wxT("csound"));
  if (m_output.IsOk()) {
    args.Add(wxT("-o"));
    args.Add(mkrel(base, m_output).GetFullPath());
  }
  if (m_log.IsOk()) {
    args.Add(wxT("-O"));
    args.Add(mkrel(base, m_log).GetFullPath());
  }
  switch (m_format) {
  case AIFF_FORMAT:
    args.Add(wxT("-A"));
    break;
  case WAVE_FORMAT:
    args.Add(wxT("-W"));
    break;
  case IRCAM_FORMAT:
    args.Add(wxT("-J"));
    break;
  case RAW_FORMAT:
    args.Add(wxT("-h"));
    break;
  case NO_FORMAT:
    break;
  }
  switch (m_output_size) {
  case UNSIGNED_EIGHT_SIZE:
    args.Add(wxT("-8"));
    break;
  case SIGNED_EIGHT_SIZE:
    args.Add(wxT("-c"));
    break;
  case SHORT_SIZE:
    args.Add(wxT("-s"));
    break;
  case TWENTY_FOUR_SIZE:
    args.Add(wxT("-3"));
    break;
  case THIRTY_TWO_SIZE:
    args.Add(wxT("-l"));
    break;
  case FLOAT_SIZE:
    args.Add(wxT("-f"));
    break;
  case NO_SIZE:
    break;
  }
  switch (m_graphics) {
  case ASCII_GRAPHICS:
    args.Add(wxT("-g"));
    break;
  case POSTSCRIPT_GRAPHICS:
    args.Add(wxT("-G"));
    break;
  case NO_GRAPHICS:
    args.Add(wxT("-d"));
    break;
  case FULL_GRAPHICS:
    break;
  }
  if (m_no_peak_chunks)
    args.Add(wxT("-K"));
  if (m_play_after_rendering)
    args.Add(wxT("-p"));
  if (m_verbose)
    args.Add(wxT("-v"));
  if (m_cscore)
    args.Add(wxT("-C"));
  if (m_init_time_only)
    args.Add(wxT("-I"));
  if (m_no_sound)
    args.Add(wxT("-n"));
  if (m_notify)
    args.Add(wxT("-N"));
  if (m_list_opcodes)
    args.Add(wxT("-z"));
  if (m_rewrite_header)
    args.Add(wxT("-R"));
  if (!m_message_level.IsEmpty()) {
    args.Add(wxT("-m"));
    args.Add(m_message_level);
  }
  if (m_orchestra.IsOk()) {
    args.Add(mkrel(base, m_orchestra).GetFullPath());
    if (m_score.IsOk())
      args.Add(mkrel(base, m_score).GetFullPath());
  }
  return args;
}

// Command line parsing

namespace {
  const wxCmdLineEntryDesc arg_desc[] =
  {
    { wxCMD_LINE_OPTION, wxT("o"), wxT("output"),
      wxT("sound output filename") },
    { wxCMD_LINE_OPTION, wxT("O"), wxT("logfile"),
      wxT("log output to file") },
    // Formats
    { wxCMD_LINE_SWITCH, wxT("A"), wxT("aiff"), wxT("set AIFF format") },
    { wxCMD_LINE_SWITCH, wxT("W"), wxT("wave"), wxT("set WAV format") },
    { wxCMD_LINE_SWITCH, wxT("J"), wxT("ircam"), wxT("set IRCAM format") },
    { wxCMD_LINE_SWITCH, wxT("h"), wxT("noheader"), wxT("raw format") },
    // Sizes
    { wxCMD_LINE_SWITCH, wxT("s"), wxT("short"),
      wxT("short int sound samples") },
    { wxCMD_LINE_SWITCH, wxT("8"), wxT("uchar"),
      wxT("unsigned char sound samples") },
    { wxCMD_LINE_SWITCH, wxT("c"), wxT("char"),
      wxT("signed char sound samples") },
    { wxCMD_LINE_SWITCH, wxT("3"), wxT("24bit"),
      wxT("24-bit sound samples") },
    { wxCMD_LINE_SWITCH, wxT("l"), wxT("long"),
      wxT("long int sound samples") },
    { wxCMD_LINE_SWITCH, wxT("f"), wxT("float"),
      wxT("float sound samples") },
    // Graphics
    { wxCMD_LINE_SWITCH, wxT("d"), wxT("nodisplays"),
      wxT("suppress all displays") },
    { wxCMD_LINE_SWITCH, wxT("g"), wxT("asciidisplay"),
      wxT("suppress graphics, use ascii displays") },
    { wxCMD_LINE_SWITCH, wxT("G"), wxT("postscriptdisplay"),
      wxT("suppress graphics, use Postscript displays") },
    // Other
    { wxCMD_LINE_SWITCH, wxT("K"), wxT("nopeaks"),
      wxT("do not write peak information") },
    { wxCMD_LINE_SWITCH, wxT("p"), wxT("play-on-end"),
      wxT("play after rendering") },
    { wxCMD_LINE_SWITCH, wxT("v"), wxT("verbose"),
      wxT("verbose orchestra translation") },
    { wxCMD_LINE_SWITCH, wxT("C"), wxT("cscore"),
      wxT("use Cscore processing of scorefile") },
    { wxCMD_LINE_SWITCH, wxT("I"), wxT("i-only"),
      wxT("I-time only orchestra run") },
    { wxCMD_LINE_SWITCH, wxT("n"), wxT("nosound"),
      wxT("no sound onto disk or device") },
    { wxCMD_LINE_SWITCH, wxT("N"), wxT("notify"),
      wxT("notify (ring the bell) when score or miditrack is done") },
    { wxCMD_LINE_SWITCH, wxT("R"), wxT("rewrite"),
      wxT("continually rewrite header while writing soundfile (WAV/AIFF)") },
    { wxCMD_LINE_SWITCH, wxT("z"), wxT("list-opcodes"),
      wxT("list opcodes in this version") },
    { wxCMD_LINE_OPTION, wxT("m"), wxT("messagelevel"),
      wxT("message level. Sum of: 1=note amps, 2=out-of-range msg, 4=warnings") },
    // Files
    { wxCMD_LINE_PARAM,  NULL, NULL, wxT("input file"), wxCMD_LINE_VAL_STRING,
      wxCMD_LINE_PARAM_OPTIONAL | wxCMD_LINE_PARAM_MULTIPLE },
    // End
    { wxCMD_LINE_NONE }
  };
}

// Generate an absolute file name
inline wxFileName mkabs(const wxFileName& base, const wxString& file_name)
{
  wxFileName name(file_name);
  if (base.IsOk())
    name.MakeAbsolute(base.GetPath(wxPATH_GET_SEPARATOR));
  else
    name.MakeAbsolute();
  return name;
}

int Args::parse(const wxFileName& base, const wxArrayString& args)
{
  int argc = args.GetCount();
  wxChar *argv[argc + 1];
  for (int i = 0; i < argc; i++)
    argv[i] = const_cast<wxChar *>(args[i].c_str());
  argv[argc] = 0;
  wxCmdLineParser parser(arg_desc, argc, argv);
  parser.SetSwitchChars(wxT("-"));
  parser.DisableLongOptions();
  int rc = parser.Parse();
  if (rc != 0)
    return rc;
  clear();

  wxString str;
  if (parser.Found(wxT("o"), &str))
    m_output = mkabs(base, str);
  if (parser.Found(wxT("O"), &str))
    m_log = mkabs(base, str);

  if (parser.Found(wxT("A")))
    m_format = AIFF_FORMAT;
  else if (parser.Found(wxT("W")))
    m_format = WAVE_FORMAT;
  else if (parser.Found(wxT("J")))
    m_format = IRCAM_FORMAT;
  else if (parser.Found(wxT("h")))
    m_format = RAW_FORMAT;

  if (parser.Found(wxT("s")))
    m_output_size = SHORT_SIZE;
  else if (parser.Found(wxT("8")))
    m_output_size = UNSIGNED_EIGHT_SIZE;
  else if (parser.Found(wxT("c")))
    m_output_size = SIGNED_EIGHT_SIZE;
  else if (parser.Found(wxT("3")))
    m_output_size = TWENTY_FOUR_SIZE;
  else if (parser.Found(wxT("l")))
    m_output_size = THIRTY_TWO_SIZE;
  else if (parser.Found(wxT("f")))
    m_output_size = FLOAT_SIZE;

  if (parser.Found(wxT("d")))
    m_graphics = NO_GRAPHICS;
  else if (parser.Found(wxT("G")))
    m_graphics = POSTSCRIPT_GRAPHICS;
  else if (parser.Found(wxT("g")))
    m_graphics = ASCII_GRAPHICS;

  m_no_peak_chunks = parser.Found(wxT("K"));
  m_play_after_rendering = parser.Found(wxT("p"));
  m_verbose = parser.Found(wxT("v"));
  m_cscore = parser.Found(wxT("C"));
  m_init_time_only = parser.Found(wxT("I"));
  m_no_sound = parser.Found(wxT("n"));
  m_notify = parser.Found(wxT("N"));
  m_rewrite_header = parser.Found(wxT("R"));
  m_list_opcodes = parser.Found(wxT("z"));

  parser.Found(wxT("m"), &m_message_level);

  size_t count = parser.GetParamCount();
  if (count > 0) {
    m_orchestra = mkabs(base, parser.GetParam(0));
    if (count > 1)
      m_score = mkabs(base, parser.GetParam(1));
  }
  update();
  return rc;
}

void Args::usage()
{
  wxCmdLineParser parser(arg_desc);
  parser.SetSwitchChars(wxT("-"));
  parser.DisableLongOptions();
  parser.Usage();
}

// Array string quoting and unquoting.  Used to store array strings in
// files.

// Space and tab delimit arguments.  Both are called a blank.
// Back slash quotes the next character.
// End of string, carriage return, and newline end input.

inline bool is_blank(int c)
{
  return c == ' ' || c == '\t';
}

inline bool is_eol(int c)
{
  return c == '\n' || c == '\r';
}

wxArrayString dequote(const wxString& line)
{
  wxArrayString out;
  size_t n = line.Len();
  wxChar c;
  size_t i = 0;

  // Remove leading blanks

  do {
    if (i >= n)
      return out;
    c = line.GetChar(i++);
    if (is_eol(c))
      return out;
  }
  while (is_blank(c));

  wxString item;
  while (1) {			// Non-blank, non-eol char here
    bool blank = FALSE;
    if (c == '\\') {
      if (i >= n) {		// This is really an error condition
	if (!item.IsEmpty())
	  out.Add(item);
	return out;
      }
      c = line.GetChar(i++);
      item.Append(c);
    }
    else
      item.Append(c);

    // This loop compresses multiple blanks into one.
    while (1) {
      if (i >= n) {
	out.Add(item);
	return out;
      }
      c = line.GetChar(i++);
      if (is_eol(c)) {
	out.Add(item);
	return out;
      }
      if (is_blank(c))
	blank = TRUE;
      else
	break;
    }

    if (blank) {		// New item found
      out.Add(item);
      item.Clear();
      blank = FALSE;
    }
  }
}

// The characters that need not be quoted.
inline bool is_plain(int c)
{
  switch (c) {
  case '@':
  case '_':
  case '-':
  case '+':
  case '=':
  case ':':
  case ',':
  case '.':
  case '/':
    return TRUE;
  default:
    return isalnum(c);
  }
}

// Add an item to the output
inline void quote(wxString& out, const wxString& item)
{
  size_t n = item.Len();
  for (size_t i = 0; i < n; i++) {
    wxChar c = item.GetChar(i);
    if (!is_plain(c))
      out.Append('\\');
    out.Append(c);
  }
}

wxString enquote(const wxArrayString& args)
{
  wxString out;
  size_t n = args.GetCount();
  if (n <= 0)
    return out;
  quote(out, args.Item(0));

  for (size_t i = 1; i < n; i++) {
    out.Append(wxT(" "));
    quote(out, args.Item(i));
  }
  return out;
}
