/*
 * Canvas on which to draw curves
 *
 * Copyright (c) 2003 by John D. Ramsdell. All rights reserved.
 *
 * This software is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// Draws a set of curves using the algorithm in winX11.c.  The curves
// are extracted from the synchronized table whenever new curves
// become available.

#include <wx/wxprec.h>

#ifndef WX_PRECOMP
#include <wx/window.h>
#include <wx/panel.h>
#include <wx/sizer.h>
#include <wx/dc.h>
#include <wx/combobox.h>
#endif

#include "curve.hpp"
#include "table.hpp"
#include "canvas.hpp"

const int id_canvas_combo_box = 9909;
const int gutter_height_percent = 20;
const int border_width_percent = 10;

// The window used to draw a curve

class Plot_curve: public wxWindow
{
public:
  Plot_curve(wxWindow *);
  void on_paint(wxPaintEvent&);
  void do_paint();
  void set_curve(Curve *);
private:
  Curve *m_curve;
  void paint(wxDC&);
  void paint_curve(wxDC&, Curve *);
  DECLARE_EVENT_TABLE()
};

BEGIN_EVENT_TABLE(Plot_curve, wxWindow)
  EVT_PAINT(Plot_curve::on_paint)
END_EVENT_TABLE()

Plot_curve::Plot_curve(wxWindow *parent)
  : wxWindow(parent, -1), m_curve(0)
{
  SetBackgroundColour(*wxWHITE);
}

void Plot_curve::on_paint(wxPaintEvent& event)
{
  wxPaintDC dc(this);
  paint(dc);
}

void Plot_curve::do_paint()
{
  wxClientDC dc(this);
  paint(dc);
}

void Plot_curve::set_curve(Curve *curve)
{
  m_curve = curve;
  do_paint();
}

void Plot_curve::paint(wxDC& dc)
{
  PrepareDC(dc);		// What is this for?
  dc.BeginDrawing();

  // Erase old curve
  dc.Clear();

  if (m_curve)
    paint_curve(dc, m_curve);

  dc.EndDrawing();
}

// The algorithm used here is taken from winX11.c
void Plot_curve::paint_curve(wxDC& dc, Curve *curve)
{
  wxCoord win_x = 0;
  wxCoord win_y = 0;
  wxCoord win_w;
  wxCoord win_h;
  dc.GetSize(&win_w, &win_h);
  wxCoord gra_x = (win_w * border_width_percent) / 100;
  wxCoord gra_y = (win_h * gutter_height_percent) / 100;
  wxCoord gra_w = win_w - 2 * gra_x;
  wxCoord gra_h = win_h - 2 * gra_y;
  wxCoord y_axis = 0;
  if (gra_w <= 0)
    return;
  if (gra_h <= 0)
    return;

  Polarity pol = curve->get_polarity();
  switch (pol) {
  case POLARITY_BIPOL:
    y_axis = gra_y + gra_h/2;
    break;
  case POLARITY_NEGPOL:
    y_axis = gra_y;
    break;
  default:
    y_axis = gra_y + gra_h;
    break;
  }

  int npts = curve->get_size();
  if (npts <= 1)
    return;
  float x_scale = static_cast<float>(gra_w) / static_cast<float>(npts - 1);
  float y_scale = static_cast<float>(gra_h) / curve->get_y_scale();
  if (pol == POLARITY_BIPOL)
    y_scale /= 2.0;

  wxPoint points[npts];
  float *xpts = curve->get_data();

  for (int i = 0; i < npts; i++) {
    points[i].x = gra_x + static_cast<wxCoord>(i * x_scale);
    points[i].y = y_axis - static_cast<wxCoord>(xpts[i] * y_scale);
  }

  dc.DrawLines(npts, points);

  // Draw axises

  dc.DrawLine(gra_x, y_axis, gra_x + gra_w, y_axis);
  dc.DrawLine(gra_x, gra_y, gra_x, gra_y + gra_h);

  if (curve->is_divider_dotted()) {
    dc.SetPen(*wxBLACK_DASHED_PEN);
    dc.DrawLine(win_x + win_x/2, win_y + gutter_height_percent * 100,
		win_x + win_w/2, win_y + win_h - gutter_height_percent * 100);
    dc.SetPen(*wxBLACK_PEN);
  }
}

// The panel used to select curves to draw and draw them.

BEGIN_EVENT_TABLE(Canvas, wxPanel)
  EVT_PAINT(Canvas::on_paint)
  EVT_COMBOBOX(id_canvas_combo_box, Canvas::on_select)
END_EVENT_TABLE()

Canvas::Canvas(Table *table,
	       wxWindow *parent,
	       wxWindowID id,
	       const wxPoint& pos,
	       const wxSize& size,
	       long style,
	       const wxString& name)
  : wxPanel(parent, id, pos, size, style, name), m_table(table)
{
  m_list.DeleteContents(TRUE);
  SetBackgroundColour(*wxLIGHT_GREY);
  m_box = new wxComboBox(this, id_canvas_combo_box);
  m_plot = new Plot_curve(this);
  wxBoxSizer *sizer = new wxBoxSizer(wxVERTICAL);
  sizer->Add(m_box, 0, wxALIGN_CENTRE | wxALL, 2);
  sizer->Add(m_plot, 1, wxEXPAND | wxALL, 2);
  SetSizer(sizer);
  sizer->SetSizeHints(this);
}

void Canvas::on_paint(wxPaintEvent& event)
{
  wxPaintDC dc(this);		// dc required for Windows
  check_curve_list();		// even though it's not used
  m_plot->on_paint(event);
}

// Paint outside a paint event
void Canvas::do_paint()
{
  check_curve_list();
  m_plot->do_paint();
}

void Canvas::check_curve_list()
{
  // Ensure curve list is up-to-date.
  if (!m_table->have_curves_been_inserted()) {
    m_list.Clear();
    m_box->Clear();
    m_plot->set_curve(0);
    m_table->insert_curves(m_list);
    Curve_list::Node *node = m_list.GetFirst();
    if (node) {
      m_plot->set_curve(node->GetData());
      for(; node; node = node->GetNext()) {
	Curve *data = node->GetData();
	m_box->Append(data->get_caption());
      }
    }
  }
}

void Canvas::on_select(wxCommandEvent& event)
{
  Curve *curve = 0;
  wxString caption = m_box->GetStringSelection();
  Curve_list::Node *node = m_list.GetFirst();
  for(; node; node = node->GetNext()) {
    Curve *data = node->GetData();
    if (caption.IsSameAs(data->get_caption())) {
      curve = data;
      break;
    }
  }
  m_plot->set_curve(curve);
}
