/*
 * Csound synthesizer
 *
 * Copyright (c) 2003 by John D. Ramsdell. All rights reserved.
 *
 * This software is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdarg.h>

#include <wx/wxprec.h>

#ifndef WX_PRECOMP
#include <wx/string.h>
#include <wx/app.h>
#endif

#include <csound/csound.h>
#include "synthesizer.hpp"
#include "curve.hpp"
#include "table.hpp"
#include "main.hpp"

namespace {

#if defined CHECK_FOR_ASCII
  inline bool is_ascii(const char *buf)
  {
    for (;;) {
      char ch = *buf++;
      if (!ch)
	return TRUE;
      if (ch < 0 || ch >= 128)
	return FALSE;
    }
  }
#endif

  const int buf_size = 1 << 10;

  char buf[buf_size];

  int last = 0;			// Marks end of buffer

  // Does char array contain newline character
  inline bool is_terminated(const char *buf)
  {
    for (;;) {
      char ch = *buf++;
      if (!ch)
	return FALSE;
      if (ch == '\n')
	return TRUE;
    }
  }
}

extern "C" {

  // Sends Csound messages from the worker thread to the main thread.
  // Accumulates text in buf until a newline character appears.
  void wxcsound_message(void *csound, const char *format, va_list args)
  {
    const char *msg = 0;
    int n = vsnprintf(buf + last, buf_size - last, format, args);
    if (n == 0)
      return;
    if (n < 0)
      msg = "Internal error: Csound print failed\n";
    else if (n >= buf_size - last)
      msg = "Internal error: Csound output too big\n";
#if defined CHECK_FOR_ASCII
    else if (!is_ascii(buf))
      msg = "Internal error: Non-ASCII from Csound\n";
#endif
    if (msg) {			// On internal error, print message.
      last = 0;
      wxGetApp().post_transcript_text(wxString(msg, wxConvUTF8));
    }
    else if (last + n > (buf_size >> 1) || is_terminated(buf + last)) {
      last = 0;			// When newline or buf too full
      wxGetApp().post_transcript_text(wxString(buf, wxConvUTF8));
    }
    else			// Save text until next time
      last += n;
  }

  void csoundMessage0(const char *format, ...)
  {
    va_list args;
    va_start(args, format);
    wxcsound_message(0, format, args);
    va_end(args);
  }

  // Sends graphics curves to the main thread

  void wxcsound_make_graph(void *csound, WINDAT *windat, char *name)
  {
  }

  // Create a Curve here and send it to the main application.
  // The main application is responsible for deleting the curve.
  void wxcsound_draw_graph(void *csound, WINDAT *windat)
  {
    windat->caption[CAPSIZE - 1] = 0; // Just in case...
    Polarity polarity;
    // translate polarities and hope the definition in Csound doesn't change.
    switch (windat->polarity) {
    case NEGPOL:
      polarity = POLARITY_NEGPOL;
      break;
    case POSPOL:
      polarity = POLARITY_POSPOL;
      break;
    case BIPOL:
      polarity = POLARITY_BIPOL;
      break;
    default:
      polarity = POLARITY_NOPOL;
    }
    Curve *curve
      = new Curve(windat->fdata,
		  windat->npts,
		  wxString(windat->caption, wxConvUTF8),
		  polarity,
		  windat->max,
		  windat->min,
		  windat->absmax,
		  windat->oabsmax,
		  windat->danflag);
    Main& app = wxGetApp();
    app.post_plot_curve(curve);
  }

  void wxcsound_kill_graph(void *csound, WINDAT *windat)
  {
  }

  int wxcsound_exit_graph(void* csound)
  {
    return 0;
  }
}

Synthesizer::Synthesizer()
  : m_csound(csoundCreate(this))
{
  csoundSetMessageCallback(m_csound, wxcsound_message);
  csoundSetIsGraphable(m_csound, TRUE);
  csoundSetMakeGraphCallback(m_csound, wxcsound_make_graph);
  csoundSetDrawGraphCallback(m_csound, wxcsound_draw_graph);
  csoundSetKillGraphCallback(m_csound, wxcsound_kill_graph);
  csoundSetExitGraphCallback(m_csound, wxcsound_exit_graph);
}

Synthesizer::~Synthesizer()
{
  csoundDestroy(m_csound);
}

void Synthesizer::set_yield_callback(yield_callback_t callback)
{
  csoundSetYieldCallback(m_csound, callback);
}

int Synthesizer::perform(int argc, char **argv)
{
  Main& app = wxGetApp();
  app.post_clear_plots();
  int rc = csoundPerform(m_csound, argc, argv);
  csoundCleanup(m_csound);
  csoundReset(m_csound);
  return rc;
}

wxString Synthesizer::get_version()
{
  int version = csoundGetVersion();
  int major = version / 100;
  int minor = version % 100;
  wxString string;
  string.Printf(wxT("%d.%2d"), major, minor);
  return string;
}
